--Many of the connected structures that we need to describe in digital desing
--have a regular patterned structure. In instantiating a set of components
--we want to connect them according to a simple computational rule. The rule
--governs the shape of the structure. So the rule is not an algorithm governing
--behavior such a process.
--
--VHDL provides 'generate' statements to support the designer's need to create
--regular structure patterns. Unfortunately, our textbook includes no examples
--and no description of the generate statements. I will provide the needed 
--material here.

--Consider a typical example. We wish to create a component that is a 
--shift register of length n. We have available as components, D fliplops, DFF.

--Here is a reasonable design, using generate statements.

--entity shift_reg is
--       generic (n : natural);
--       port    ( din, clk : bit
--                 dout     : inout bit);
--end shift_reg;
--
--
--architecture struc of shift_reg is
--
--   component dff port( d, clock: bit; dout : out bit); end component;
-- 
--   signal q_int : bit_vector(0 to n-1);

--begin
--
--   L1: for i in 1 to n-1 generate
--        L2: dff port map (q_int(i-1), clk, q_int(i));
--       end generate;
--
--   L3: dff port map(din,clk,q_int(0));
--
--   dout <= q_int(n-1);
--end struc;

--It is pretty clear what structure of dff's is being constructed. 

--Here are some rules: Every statement must be labelled. This is true
--                     for instantiations in general, but for generate
--                     statements both the loop control statement and the
--                     loop body statements need labels.
--                
--                     Clearly, the generate statements must be completely
--                     resolved BEFORE the simulation run begins. No generate
--                     statement can depend on a signal, or variable, that is
--                     part of a simulation run. It is often very tempting
--                     to violate this rule.
--
--                     The main task is to consider the subscript calculations
--                     carefully in order to achieve the desired pattern.
--
--                     You need to create internal signals for all the internal
-- 
--Now we will put the shift_reg into a legal context so we can simulate it.

entity dff is port(d,c:bit; q :inout bit); end dff;
architecture beh of dff is
begin q <= d when (c'event and c = '1') else q; end beh;

entity shift_reg is
       generic (n : natural);
       port    ( din, clk : bit;
                 dout     : inout bit);
end shift_reg;

use work.all;

architecture struc of shift_reg is

   component dff port( d, c: bit; q : inout bit); end component;
 
   signal q_int : bit_vector(0 to n-1);
begin

   L1: for i in 1 to n-1 generate
        L2: dff port map (q_int(i-1), clk, q_int(i));
       end generate;
							
   L3: dff port map(din,clk,q_int(0));

   dout <= q_int(n-1);
end struc;

use work.all;

entity test_shift_reg is end;
architecture struct of test_shift_reg is
   component shift_reg generic (n : natural);
               port    ( din, clk : bit; dout : inout bit); end component;


   signal testin, testout, clock : bit;
begin
   clock <= not clock after 5 ns;
   testin   <= not testin after 31 ns;
   L: shift_reg generic map(3) port map(testin, clock, testout);
end struct;


-- So the general form of a 'for type' generate statement is:
-- <label>: for <iteration control> generate
--               <body of generate loop>
--            end generate;
--
-- generate statements can be nested so you can get 2-dimensional
-- and 3 dimensional patterns.

-- Here are two more examples. We want to define an n-input and gate,
-- with a bit_vector input, using only 2-input and gates as components.

entity and2 is port (a,b : bit; z : out bit); end and2;
architecture beh of and2 is begin z <= a and b; end beh;

use work.all;

entity andn is generic (n : natural);
               port ( din : bit_vector(0 to n-1);
                      z : out bit);
end andn;

architecture struc1 of andn is

   component and2 port (a,b : bit; z : out bit); end component;

   signal z_int : bit_vector(0 to n - 2);
begin
-- this version works if n  is 2 or greater
   L0: and2 port map(din(0),din(1),z_int(0));

   L1: for i in 1 to n-2 generate
         L2: and2 port map (z_int(i-1),din(i+1),z_int(i));
       end generate;
   z <= z_int(n-2);
end struc1;

use work.all;

architecture struc2 of andn is

   component and2 port (a,b : bit; z : out bit); end component;
   signal z_int : bit_vector(0 to 2*n-2);
begin
 
   z_int(0 to n-1) <= din;

   L1: for i in 0 to n-2 generate
       L2: and2 port map(z_int(2*i),z_int(2*i+1),z_int(n+i));
       end generate;

   z <= z_int(2*n-2);
end struc2;

-- notice in this case we have two different structures so when we construct
-- a testbench we need to specify, at each instantiation, which structure
-- to use. This is done using 'configuration statements' of the form:
-- for <label>: <component_name> use entity <entity_name>(<architecture>);
-- as shown below.

use work.all;

entity test_andn is end test_andn;

architecture struc of test_andn is

  component andn generic (n : natural);
            port ( din : bit_vector(0 to n-1);
                   z : out bit); end component;

  use work.all;
  for L1: andn use entity work.andn(struc1);
  for L2: andn use entity work.andn(struc2);

  signal testin : bit_vector (0 to 4);
  signal result1, result2 : bit;
  
begin

   L1: andn generic map (5) port map (testin,result1);
   L2: andn generic map (5) port map (testin,result2);

   testin <= ((not testin(4)) & testin(0 to 3)) after 5 ns;

end struc;

-- there is one more type of generate statement, the 'if .. generate'
-- here is an example, an alternative structure for the shift_reg

use work.all;

architecture struc2 of shift_reg is

   component dff port( d, c: bit; q : inout bit); end component;
 
   signal q_int : bit_vector(0 to n-1);
begin

   L1: for i in 0 to n-1 generate
        LI: if (i = 0) generate 
             LI1: dff port map(din,clk,q_int(0)); end generate;
        L2: if ( I /= 0 ) generate
            L21:   dff port map (q_int(i-1), clk, q_int(i)); end generate;
       end generate;

   dout <= q_int(n-1);

end struc2;

--Notice ther is no 'else' or 'elsif' for a generate usiing the if mode
-- Notice also that the above design would not work well if n = 1.
-- struc3 is done more carefully

architecture struc3 of shift_reg is

   component dff port( d, c: bit; q : inout bit); end component;
 
   signal q_int : bit_vector(0 to n-1);
begin

   L1: if (n = 1) generate 
          L11: dff port map (din,clk,dout); end generate;
   L2: if n > 1 generate
       L3: for i in 0 to n-1 generate
           L4: if (i = 0) generate 
                L41:dff port map(din,clk,q_int(0)); end generate;
           L5: if (i = n-1) generate 
                L51:dff port map (q_int(i-1),clk, dout); end generate;
           L6: if ( I >0 ) and ((i < n-1)) generate
                L61: dff port map (q_int(i-1), clk, q_int(i)); end generate;
          end generate;
       end generate;

end struc3;

-- many of the most attractive hardware strucures are recursive in nature
-- designers liske to build big registers out of smaller ones, big adders
-- out of smaller ones, and so on. In such cases the 'internal' signals
-- are replicated with every recursion so there is a problem with name
-- scope. We avoid the problem with the 'block' statement 
-- Here is a new architecture for shift_reg

architecture struc4 of shift_reg is 

   component dff port( d, c: bit; q : inout bit); end component;
   component shift_reg generic (n : natural);
               port    ( din, clk : bit; dout : inout bit); end component;

begin

   L1: if (n = 1) generate 
        L11: dff port map (din,clk,dout); end generate;

   L2: if (n > 1) generate
       B: block
           signal temp : bit;
           begin
            L3: shift_reg generic map ( n / 2) port map ( din, clk, temp);
            L4: shift_reg generic map ( n - n/2) port map(temp,clk,dout);
          end block;
       end generate;
end struc4;        


-- Here is an example of building a big adder from small ones, recursive

entity faddr is port ( a,b,c : bit;
                       s,cout : out bit); 
end faddr;

architecture beh of faddr is
begin
   s <= a xor b xor c;
   cout <= (a and b) or (a and c) or (b and c);
end;

entity addern is generic (n : natural);
                 port ( an : bit_vector(n-1 downto 0);
                        bn : bit_vector(n-1 downto 0);
                        carry_in : bit;
                        sn : inout bit_vector(n-1 downto 0);
                        carry_out : out bit);
end addern;

architecture struc of addern is

  component faddr port ( a,b,c : bit;
                         s,cout : out bit); end component;

  component addern generic (n : natural);
                 port ( an : bit_vector(n-1 downto 0);
                        bn : bit_vector(n-1 downto 0);
                        carry_in : bit;
                        sn : inout bit_vector(n-1 downto 0);
                        carry_out: out bit); end component;
begin

   L0: if (n = 1) generate 
                   L: faddr port map (an(0),bn(0),carry_in,sn(0),carry_out); 
                  end generate;
   LN: if (n > 1) generate
    B: block
        signal carr : bit;
        begin
--          L1: addern generic map (n / 2) port map(a(((n / 2) -1) downto 0),
--                                                  b(((n / 2) -1) downto 0),
--                                                  carry_in, 
--                                                  s(((n / 2) -1) downto 0),
--                                                  carr);
          L2: addern generic map(n- (n / 2)) port map(an((n-1) downto (n / 2)),
                                                      bn((n-1) downto (n / 2)),
                                                      carr, 
                                                      sn((n-1) downto (n / 2)),
                                                      carry_out);
      end block;
   end generate;
end   struc;

--There are some examples where the recursive structures appear to be
--so superior that other methods are quite impractical. Consider the 
--design of a decoder. The decoder has n inputs and 2**n outputs. Every
-- big memory chip has decoder logic inside so decoder design is quite
-- important. Both speed and area are important.
--
-- in the following design big decoders are composed of two 1/2 size
-- decoders and a bank of and gates

entity decoder is generic (n:natural); -- the number of inputs
                  port ( addr : bit_vector(0 to (n-1));
                         sel  : out bit_vector (0 to (2**n - 1)));
end;

use work.all;

architecture struc of decoder is

   component and2 port(a,b:bit; z: out bit); end component;
 
   component decoder generic (n: natural); -- the number of inputs
                    port ( addr : bit_vector(0 to (n-1));
                           sel  : out bit_vector (0 to (2**n - 1)));
            end component;

begin
  L0: if (n=1) generate
      L1: sel <=  (not addr(0)) & addr(0);
      end generate;

 L2: if n > 1 generate
      L3: block
           signal temp1 : bit_vector(0 to 2**(n/2) - 1);
           signal temp2 : bit_vector(0 to 2**(n - (n/2)) -1);
           begin
             L4: decoder generic map (n/2)
                         port map (addr(0 to ((n/2)-1)),temp1);
             L5: decoder generic map(n- (n/2))
                         port map (addr(n/2 to n-1),temp2);
             L6: for i in 0 to 2**(n/2) -1 generate        
                 L7: for j in 0 to 2**(n-(n/2))-1 generate
                      L8: and2 port map (temp1(i),temp2(j),
                                         sel(i*(2**(n-n/2)) + j));
                     end generate;
                 end generate;
           end block;
       end generate;

end struc;


entity test_decoder is end test_decoder;

architecture struc of test_decoder is

   component decoder generic (n: natural); -- the number of inputs
                    port ( addr : bit_vector(0 to (n-1));
                           sel  : out bit_vector (0 to (2**n - 1)));
            end component;

constant n : integer := 5;
signal inval  : bit_vector(0 to n-1);
signal outval : bit_vector(0 to 2**n -1);
begin

  inval <= not inval(n-1) & inval(0 to n-2)  after 10 ns;

  L: decoder generic map (n) port map (inval, outval);
end;
